#include <mesh/MeshSmootherEscobar.hpp>

#include <algebra/TinyMatrix.hpp>
#include <algebra/TinyVector.hpp>
#include <geometry/TriangleTransformation.hpp>
#include <language/utils/InterpolateItemValue.hpp>
#include <mesh/Connectivity.hpp>
#include <mesh/ItemValueUtils.hpp>
#include <mesh/ItemValueVariant.hpp>
#include <mesh/Mesh.hpp>
#include <mesh/MeshCellZone.hpp>
#include <mesh/MeshFlatNodeBoundary.hpp>
#include <mesh/MeshLineNodeBoundary.hpp>
#include <mesh/MeshNodeBoundary.hpp>
#include <scheme/AxisBoundaryConditionDescriptor.hpp>
#include <scheme/DiscreteFunctionUtils.hpp>
#include <scheme/DiscreteFunctionVariant.hpp>
#include <scheme/FixedBoundaryConditionDescriptor.hpp>
#include <scheme/SymmetryBoundaryConditionDescriptor.hpp>
#include <utils/RandomEngine.hpp>

#include <variant>

template <size_t Dimension>
class MeshSmootherEscobarHandler::MeshSmootherEscobar
{
 private:
  using Rd               = TinyVector<Dimension>;
  using Rdxd             = TinyMatrix<Dimension>;
  using ConnectivityType = Connectivity<Dimension>;
  using MeshType         = Mesh<ConnectivityType>;

  const MeshType& m_given_mesh;
  std::shared_ptr<const IMesh> m_p_given_mesh;

  class AxisBoundaryCondition;
  class FixedBoundaryCondition;
  class SymmetryBoundaryCondition;

  static constexpr double m_min_quality = 0.2;

  using BoundaryCondition = std::variant<AxisBoundaryCondition, FixedBoundaryCondition, SymmetryBoundaryCondition>;

  using BoundaryConditionList = std::vector<BoundaryCondition>;
  BoundaryConditionList m_boundary_condition_list;

  NodeValue<const bool> m_is_fixed_node;
  NodeValue<const bool> m_is_smoothed_node;

  BoundaryConditionList
  _getBCList(const MeshType& mesh,
             const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>& bc_descriptor_list) const
  {
    BoundaryConditionList bc_list;

    for (const auto& bc_descriptor : bc_descriptor_list) {
      switch (bc_descriptor->type()) {
      case IBoundaryConditionDescriptor::Type::axis: {
        if constexpr (Dimension == 1) {
          bc_list.emplace_back(FixedBoundaryCondition{getMeshNodeBoundary(mesh, bc_descriptor->boundaryDescriptor())});
        } else {
          bc_list.emplace_back(
            AxisBoundaryCondition{getMeshLineNodeBoundary(mesh, bc_descriptor->boundaryDescriptor())});
        }
        break;
      }
      case IBoundaryConditionDescriptor::Type::symmetry: {
        bc_list.emplace_back(
          SymmetryBoundaryCondition{getMeshFlatNodeBoundary(mesh, bc_descriptor->boundaryDescriptor())});
        break;
      }
      case IBoundaryConditionDescriptor::Type::fixed: {
        bc_list.emplace_back(FixedBoundaryCondition{getMeshNodeBoundary(mesh, bc_descriptor->boundaryDescriptor())});
        break;
      }
      default: {
        std::ostringstream error_msg;
        error_msg << *bc_descriptor << " is an invalid boundary condition for mesh smoother";
        throw NormalError(error_msg.str());
      }
      }
    }

    return bc_list;
  }

  NodeValue<const bool>
  _getFixedNodes() const
  {
    NodeValue<bool> is_fixed{m_given_mesh.connectivity()};
    is_fixed.fill(false);

    for (auto&& boundary_condition : m_boundary_condition_list) {
      std::visit(
        [&](auto&& bc) {
          using BCType = std::decay_t<decltype(bc)>;
          if constexpr ((std::is_same_v<BCType, FixedBoundaryCondition>) or
                        ((Dimension == 2) and std::is_same_v<BCType, AxisBoundaryCondition>)) {
            const Array<const NodeId>& node_list = bc.nodeList();
            parallel_for(
              node_list.size(), PUGS_LAMBDA(size_t i_node) { is_fixed[node_list[i_node]] = true; });
          }
        },
        boundary_condition);
    }

    if constexpr (Dimension == 3) {
      throw NotImplementedError("treat sliding axis node kind");
    }

    synchronize(is_fixed);

    NodeValue<int> bc_number{m_given_mesh.connectivity()};
    bc_number.fill(-1);

    {
      int bc_id = 0;

      for (auto&& boundary_condition : m_boundary_condition_list) {
        std::visit(
          [&](auto&& bc) {
            using BCType = std::decay_t<decltype(bc)>;
            if constexpr (std::is_same_v<BCType, SymmetryBoundaryCondition>) {
              const Array<const NodeId>& node_list = bc.nodeList();
              parallel_for(
                node_list.size(), PUGS_LAMBDA(size_t i_node) {
                  const NodeId node_id = node_list[i_node];
                  if (not is_fixed[node_id]) {
                    if (bc_number[node_id] < 0) {
                      bc_number[node_id] = bc_id;
                    } else {
                      bc_number[node_id] = -2;   // error tag
                    }
                  }
                });
            }
          },
          boundary_condition);
        bc_id++;
      }

      if (min(bc_number) < -1) {
        throw NormalError("Smoothing boundary conditions overlap. One must precise fixed points and smoothing axis");
      }
    }

    bool missing_bc  = false;
    auto is_boundary = m_given_mesh.connectivity().isBoundaryNode();
    parallel_for(m_given_mesh.numberOfNodes(), [&](NodeId node_id) {
      if (is_boundary[node_id]) {
        if (not is_fixed[node_id] and (bc_number[node_id] == -1)) {
          missing_bc = true;
        }
      }
    });

    if (parallel::allReduceOr(missing_bc)) {
      throw NormalError("Some boundary conditions are missing (boundary nodes without BC)");
    }

    return is_fixed;
  }

  void
  _applyBC(const NodeValue<const Rd>& old_xr, NodeValue<Rd>& new_xr) const
  {
    for (auto&& boundary_condition : m_boundary_condition_list) {
      std::visit(
        [&](auto&& bc) {
          using BCType = std::decay_t<decltype(bc)>;
          if constexpr (std::is_same_v<BCType, SymmetryBoundaryCondition>) {
            const Rd& n = bc.outgoingNormal();

            const auto node_list = bc.nodeList();

            const ConnectivityType& connectivity = m_given_mesh.connectivity();

            auto is_boundary_node = connectivity.isBoundaryNode();
            auto node_is_owned    = connectivity.nodeIsOwned();

            if constexpr (Dimension == 2) {
              const TinyVector<Dimension> t{-n[1], n[0]};

              auto cell_to_node_matrix        = connectivity.cellToNodeMatrix();
              auto node_to_cell_matrix        = connectivity.nodeToCellMatrix();
              auto node_number_in_their_cells = connectivity.nodeLocalNumbersInTheirCells();
              auto node_to_edge_matrix        = connectivity.nodeToEdgeMatrix();
              auto edge_to_node_matrix        = connectivity.edgeToNodeMatrix();

              constexpr double eps = 1E-14;

              auto smooth = [=](const NodeId node_id, TinyVector<Dimension>& x) -> double {
                auto cell_list           = node_to_cell_matrix[node_id];
                auto node_number_in_cell = node_number_in_their_cells[node_id];

                const double alpha = std::acos(-1) / cell_list.size();

                const double cos_alpha     = std::cos(alpha);
                const double sin_alpha     = std::sin(alpha);
                const double inv_sin_alpha = 1. / sin_alpha;
                const double inv_tan_alpha = 1. / std::tan(alpha);

                const TinyMatrix<Dimension> W{1, cos_alpha,   //
                                              0, sin_alpha};

                const TinyMatrix<Dimension> inv_W = inverse(W);

                TinyMatrix<Dimension, Dimension> dtheta_S =
                  TinyMatrix<Dimension>{t[0], t[0] * (inv_sin_alpha - inv_tan_alpha),   //
                                        t[1], t[1] * (inv_sin_alpha - inv_tan_alpha)};

                SmallArray<TinyMatrix<Dimension>> S_list(cell_list.size());
                for (size_t i_cell = 0; i_cell < cell_list.size(); ++i_cell) {
                  const size_t i_cell_node   = node_number_in_cell[i_cell];
                  const auto cell_node_list  = cell_to_node_matrix[cell_list[i_cell]];
                  const size_t cell_nb_nodes = cell_node_list.size();

                  const TinyVector a = old_xr[cell_node_list[(i_cell_node + 1) % cell_nb_nodes]];
                  const TinyVector b = old_xr[cell_node_list[(i_cell_node + cell_nb_nodes - 1) % cell_nb_nodes]];

                  const TinyMatrix<Dimension> A{a[0] - x[0], b[0] - x[0],   //
                                                a[1] - x[1], b[1] - x[1]};

                  S_list[i_cell] = A * inv_W;
                }

                SmallArray<double> sigma_list(S_list.size());
                for (size_t i_cell = 0; i_cell < S_list.size(); ++i_cell) {
                  sigma_list[i_cell] = det(S_list[i_cell]);
                }

                double sigma_min = min(sigma_list);

                double delta = (sigma_min < eps)
                                 ? std::max(std::sqrt(eps * (eps - sigma_min)), std::sqrt(eps) * std::abs(sigma_min))
                                 : 0;

                auto frobenius = [](const TinyMatrix<Dimension>& M) { return std::sqrt(trace(transpose(M) * M)); };

                double final_f = 0;

                for (size_t i_iter = 0; i_iter < 200; ++i_iter) {
                  SmallArray<TinyMatrix<Dimension>> S_list(cell_list.size());
                  for (size_t i_cell = 0; i_cell < cell_list.size(); ++i_cell) {
                    const size_t i_cell_node   = node_number_in_cell[i_cell];
                    auto cell_node_list        = cell_to_node_matrix[cell_list[i_cell]];
                    const size_t cell_nb_nodes = cell_node_list.size();

                    const TinyVector a = old_xr[cell_node_list[(i_cell_node + 1) % cell_nb_nodes]];
                    const TinyVector b = old_xr[cell_node_list[(i_cell_node + cell_nb_nodes - 1) % cell_nb_nodes]];

                    const TinyMatrix<Dimension> A{a[0] - x[0], b[0] - x[0],   //
                                                  a[1] - x[1], b[1] - x[1]};

                    S_list[i_cell] = A * inv_W;
                  }

                  SmallArray<double> sigma_list(S_list.size());
                  for (size_t i_cell = 0; i_cell < S_list.size(); ++i_cell) {
                    sigma_list[i_cell] = det(S_list[i_cell]);
                  }

                  {
                    double current_sigma_min = min(sigma_list);
                    if (current_sigma_min < sigma_min) {
                      sigma_min = current_sigma_min;
                      delta     = (sigma_min < eps)
                                ? std::max(std::sqrt(eps * (eps - sigma_min)), std::sqrt(eps) * std::abs(sigma_min))
                                : 0;
                    }
                  }

                  double f         = 0;
                  double dtheta_f  = 0;
                  double d2theta_f = 0;

                  for (size_t i_cell = 0; i_cell < S_list.size(); ++i_cell) {
                    const double sigma            = sigma_list[i_cell];
                    const TinyMatrix<Dimension> S = S_list[i_cell];

                    const TinyMatrix<Dimension> Sigma = [&S] {
                      TinyMatrix<Dimension> transposed_comS;
                      for (size_t i = 0; i < Dimension; ++i) {
                        for (size_t j = 0; j < Dimension; ++j) {
                          transposed_comS(i, j) = cofactor(S, j, i);
                        }
                      }
                      return transposed_comS;
                    }();

                    const double S_norm      = frobenius(S);
                    const double Sigma_norm  = frobenius(Sigma);
                    const double S_norm2     = S_norm * S_norm;
                    const double Sigma_norm2 = Sigma_norm * Sigma_norm;

                    const double h = sigma + std::sqrt(sigma * sigma + 4 * delta * delta);

                    const double f_jr = S_norm * Sigma_norm / h;

                    const double dtheta_sigma = trace(Sigma * dtheta_S);

                    TinyMatrix<Dimension, Dimension> dtheta_Sigma =
                      TinyMatrix<Dimension>{t[1] * (inv_sin_alpha - inv_tan_alpha),
                                            -t[0] * (inv_sin_alpha - inv_tan_alpha),   //
                                            -t[1], t[0]};

                    const double g = trace(transpose(S) * dtheta_S) / S_norm2                 //
                                     + trace(transpose(Sigma) * dtheta_Sigma) / Sigma_norm2   //
                                     - dtheta_sigma / (h - sigma);

                    const double dtheta_f_jr = f_jr * g;
                    const double dtheta2_f_jr =
                      (trace(transpose(dtheta_S) * dtheta_S) / S_norm2                                               //
                       - 2 * trace(transpose(S) * dtheta_S) * trace(transpose(S) * dtheta_S) / (S_norm2 * S_norm2)   //
                                                                                                                     //
                       + trace(transpose(dtheta_Sigma) * dtheta_Sigma) / Sigma_norm2   // + 0
                       - 2 * trace(transpose(Sigma) * dtheta_Sigma) * trace(transpose(Sigma) * dtheta_Sigma) /
                           (Sigma_norm2 * Sigma_norm2)   //
                       //
                       - 2 * trace(dtheta_Sigma * dtheta_S) / (h - sigma)                     //
                       + 2 * sigma / (std::pow(h - sigma, 3)) * dtheta_sigma * dtheta_sigma   //
                       + g * g) *
                      f_jr;

                    f += f_jr;
                    dtheta_f += dtheta_f_jr;
                    d2theta_f += dtheta2_f_jr;
                  }
                  if (std::abs(dtheta_f) < 1E-12) {
                    break;
                  }

                  const double max_edge_lenght = [=] {
                    auto node_edge_list = node_to_edge_matrix[node_id];
                    double max_lenght   = 0;
                    for (size_t i_edge = 0; i_edge < node_edge_list.size(); ++i_edge) {
                      const EdgeId edge_id            = node_edge_list[i_edge];
                      auto edge_node_list             = edge_to_node_matrix[edge_id];
                      const TinyVector<Dimension>& x0 = old_xr[edge_node_list[0]];
                      const TinyVector<Dimension>& x1 = old_xr[edge_node_list[1]];
                      const double lenght             = l2Norm(x0 - x1);
                      if (lenght > max_lenght) {
                        max_lenght = lenght;
                      }
                    }
                    return max_lenght;
                  }();

                  const double delta_x_candidate = dtheta_f / std::abs(d2theta_f);
                  const double delta_x_norm      = std::abs(delta_x_candidate);

                  if (delta_x_norm > 0.3 * max_edge_lenght) {
                    x += (0.3 * max_edge_lenght / delta_x_norm) * delta_x_candidate * t;
                  } else {
                    x += delta_x_candidate * t;
                  }

                  if (delta_x_norm < 1E-2 * max_edge_lenght) {
                    break;
                  }

                  final_f = f;
                }
                return final_f;
              };

              parallel_for(
                node_list.size(), PUGS_LAMBDA(const size_t i_node) {
                  const NodeId node_id = node_list[i_node];
                  if (m_is_smoothed_node[node_id] and node_is_owned[node_id]) {
                    smooth(node_id, new_xr[node_id]);
                    new_xr[node_id] = 0.5 * (new_xr[node_id] + old_xr[node_id]);
                  }
                });

            } else {
              throw NotImplementedError("Dimension != 2");
            }

          } else if constexpr (std::is_same_v<BCType, AxisBoundaryCondition>) {
            if constexpr (Dimension > 1) {
              throw NotImplementedError("Escobar: axis boundary conditions");
            } else {
              throw UnexpectedError("AxisBoundaryCondition make no sense in dimension 1");
            }

          } else if constexpr (std::is_same_v<BCType, FixedBoundaryCondition>) {
            const Array<const NodeId>& node_list = bc.nodeList();
            parallel_for(
              node_list.size(), PUGS_LAMBDA(const size_t i_node) {
                const NodeId node_id = node_list[i_node];
                new_xr[node_id]      = old_xr[node_id];
              });

          } else {
            throw UnexpectedError("invalid boundary condition type");
          }
        },
        boundary_condition);
    }
  }

  CellValue<const int>
  _getNonConvexCellTag(const NodeValue<const Rd>& xr) const
  {
    CellValue<int> non_convex_cell_tag{m_given_mesh.connectivity()};

    auto cell_to_node_matrix = m_given_mesh.connectivity().cellToNodeMatrix();

    if constexpr (Dimension == 2) {
      parallel_for(
        m_given_mesh.numberOfCells(), PUGS_LAMBDA(const CellId cell_id) {
          auto cell_node_list = cell_to_node_matrix[cell_id];
          bool is_convex      = true;
          for (size_t i_node = 0; i_node < cell_node_list.size(); ++i_node) {
            const TriangleTransformation<Dimension> T{xr[cell_node_list[i_node]],
                                                      xr[cell_node_list[(i_node + 1) % cell_node_list.size()]],
                                                      xr[cell_node_list[(i_node + 2) % cell_node_list.size()]]};
            if (T.jacobianDeterminant() <= 0) {
              is_convex = false;
              break;
            }
          }
          non_convex_cell_tag[cell_id] = (is_convex == false);
        });
    } else {
      throw NotImplementedError("non convex cell calculation in dimension != 2");
    }

    return non_convex_cell_tag;
  }

  [[deprecated]] NodeValue<const bool>
  _getSmoothedNode(const CellValue<const int>& non_convex_cell_tag, const NodeValue<const bool>& m_is_fixed_node) const
  {
    auto node_to_cell_matrix = m_given_mesh.connectivity().nodeToCellMatrix();
    NodeValue<bool> is_smoothed{m_given_mesh.connectivity()};
    is_smoothed.fill(false);
    parallel_for(
      m_given_mesh.numberOfNodes(), PUGS_LAMBDA(const NodeId node_id) {
        if (not m_is_fixed_node[node_id]) {
          auto node_cell_list = node_to_cell_matrix[node_id];
          for (size_t i_cell = 0; i_cell < node_cell_list.size(); ++i_cell) {
            const CellId cell_id = node_cell_list[i_cell];
            if (non_convex_cell_tag[cell_id]) {
              is_smoothed[node_id] = true;
              break;
            }
          }
        }
      });

    return is_smoothed;
  }

  NodeValue<const bool>
  _getSmoothedNode(const NodeValue<const double>& node_quality, const NodeValue<const bool>& m_is_fixed_node) const
  {
    auto node_to_cell_matrix = m_given_mesh.connectivity().nodeToCellMatrix();
    NodeValue<bool> is_smoothed{m_given_mesh.connectivity()};
    is_smoothed.fill(false);
    parallel_for(
      m_given_mesh.numberOfNodes(), PUGS_LAMBDA(const NodeId node_id) {
        is_smoothed[node_id] = ((not m_is_fixed_node[node_id]) and (node_quality[node_id] < m_min_quality));
      });

    return is_smoothed;
  }

  void
  get_f_grad_hess(const NodeValue<const TinyVector<Dimension>>& old_xr,
                  double& sigma_min,
                  double& delta,
                  const NodeId node_id,
                  const TinyVector<Dimension>& x,
                  double& f,
                  TinyVector<Dimension>& f_gradient,
                  TinyMatrix<Dimension>& f_hessian) const
  {
    constexpr bool multi_connection = true;
    constexpr double eps            = 1E-14;

    auto frobenius = [](const TinyMatrix<Dimension>& M) { return std::sqrt(trace(transpose(M) * M)); };

    const ConnectivityType& connectivity = m_given_mesh.connectivity();

    auto cell_to_node_matrix        = connectivity.cellToNodeMatrix();
    auto node_to_cell_matrix        = connectivity.nodeToCellMatrix();
    auto node_number_in_their_cells = connectivity.nodeLocalNumbersInTheirCells();

    auto cell_list           = node_to_cell_matrix[node_id];
    auto node_number_in_cell = node_number_in_their_cells[node_id];

    const double alpha         = 2 * std::acos(-1) / cell_list.size();
    const double sin_alpha     = std::sin(alpha);
    const double cos_alpha     = std::cos(alpha);
    const double inv_sin_alpha = 1. / sin_alpha;
    const double inv_tan_alpha = cos_alpha / sin_alpha;

    const TinyMatrix<Dimension> W{1, cos_alpha,   //
                                  0, sin_alpha};

    const TinyMatrix<Dimension> inv_W = inverse(W);

    std::array<TinyMatrix<Dimension>, Dimension>                                //
      S_gradient = {TinyMatrix<Dimension>{-1, -inv_sin_alpha + inv_tan_alpha,   //
                                          +0, +0},
                    TinyMatrix<Dimension>{+0, +0,   //
                                          -1, -inv_sin_alpha + inv_tan_alpha}};

    SmallArray<TinyMatrix<Dimension>> S_list(cell_list.size());
    for (size_t i_cell = 0; i_cell < cell_list.size(); ++i_cell) {
      const size_t i_cell_node   = node_number_in_cell[i_cell];
      auto cell_node_list        = cell_to_node_matrix[cell_list[i_cell]];
      const size_t cell_nb_nodes = cell_node_list.size();

      const TinyVector a = old_xr[cell_node_list[(i_cell_node + 1) % cell_nb_nodes]];
      const TinyVector b = old_xr[cell_node_list[(i_cell_node + cell_nb_nodes - 1) % cell_nb_nodes]];

      const TinyMatrix<Dimension> A{a[0] - x[0], b[0] - x[0],   //
                                    a[1] - x[1], b[1] - x[1]};

      S_list[i_cell] = A * inv_W;
    }

    SmallArray<double> sigma_list(S_list.size());
    for (size_t i_cell = 0; i_cell < S_list.size(); ++i_cell) {
      sigma_list[i_cell] = det(S_list[i_cell]);
    }

    {
      double current_sigma_min = std::min(min(sigma_list), delta);

      if constexpr (multi_connection) {
        size_t nb_medium_range_simplices = 0;
        for (size_t i_cell = 0; i_cell < cell_list.size(); ++i_cell) {
          const CellId cell_id = cell_list[i_cell];
          nb_medium_range_simplices += 2 * (cell_to_node_matrix[cell_id].size() - 3);
        }

        if (nb_medium_range_simplices > 0) {
          SmallArray<double> medium_range_sigma_list(nb_medium_range_simplices);

          for (size_t i_cell = 0; i_cell < cell_list.size(); ++i_cell) {
            const CellId cell_id       = cell_list[i_cell];
            const size_t i_cell_node   = node_number_in_cell[i_cell];
            auto cell_node_list        = cell_to_node_matrix[cell_list[i_cell]];
            const size_t cell_nb_nodes = cell_node_list.size();

            const double mr_alpha     = 2 * std::acos(-1) / ((cell_nb_nodes - 2) * (cell_list.size()));
            const double mr_sin_alpha = std::sin(mr_alpha);
            const double mr_cos_alpha = std::cos(mr_alpha);

            const TinyMatrix<Dimension> mr_W{1, mr_cos_alpha,   //
                                             0, mr_sin_alpha};

            const TinyMatrix mr_inv_W = inverse(mr_W);

            {   // first simplex
              const TinyVector a = old_xr[cell_node_list[(i_cell_node + 1) % cell_nb_nodes]];
              const TinyVector b = old_xr[cell_node_list[(i_cell_node + 2) % cell_nb_nodes]];

              const TinyMatrix<Dimension> mr_A{a[0] - x[0], b[0] - x[0],   //
                                               a[1] - x[1], b[1] - x[1]};

              const TinyMatrix mr_S = mr_A * mr_inv_W;
              const double mr_sigma = det(mr_S);
              if (mr_sigma < current_sigma_min) {
                current_sigma_min = mr_sigma;
              }
            }

            {   // second simplex
              const TinyVector a = old_xr[cell_node_list[(i_cell_node + cell_nb_nodes - 2) % cell_nb_nodes]];
              const TinyVector b = old_xr[cell_node_list[(i_cell_node + cell_nb_nodes - 1) % cell_nb_nodes]];

              const TinyMatrix<Dimension> mr_A{a[0] - x[0], b[0] - x[0],   //
                                               a[1] - x[1], b[1] - x[1]};

              const TinyMatrix mr_S = mr_A * mr_inv_W;
              const double mr_sigma = det(mr_S);
              if (mr_sigma < current_sigma_min) {
                current_sigma_min = mr_sigma;
              }
            }
          }
        }
      }

      if (current_sigma_min < sigma_min) {
        sigma_min = current_sigma_min;
        delta =
          (sigma_min < eps) ? std::max(std::sqrt(eps * (eps - sigma_min)), std::sqrt(eps) * std::abs(sigma_min)) : 0;
      }
    }

    auto escobarSSigma = [&frobenius, delta](const double sigma, const TinyMatrix<Dimension>& S,
                                             const TinyVector<Dimension>& sigma_gradient,
                                             const std::array<TinyMatrix<Dimension>, Dimension>& S_gradient,
                                             const std::array<TinyMatrix<Dimension>, Dimension>& Sigma_gradient,
                                             double& f, TinyVector<Dimension>& f_gradient,
                                             TinyMatrix<Dimension>& f_hessian) {
      const TinyMatrix<Dimension> Sigma = [&S] {
        TinyMatrix<Dimension> transposed_comS;
        for (size_t i = 0; i < Dimension; ++i) {
          for (size_t j = 0; j < Dimension; ++j) {
            transposed_comS(i, j) = cofactor(S, j, i);
          }
        }
        return transposed_comS;
      }();

      const double S_norm      = frobenius(S);
      const double Sigma_norm  = frobenius(Sigma);
      const double S_norm2     = S_norm * S_norm;
      const double Sigma_norm2 = Sigma_norm * Sigma_norm;

      double h = sigma + std::sqrt(sigma * sigma + 4 * delta * delta);

      const double f_jr = S_norm * Sigma_norm / h;

      TinyVector<Dimension> g{trace(transpose(S) * S_gradient[0]) / S_norm2                   //
                                + trace(transpose(Sigma) * Sigma_gradient[0]) / Sigma_norm2   //
                                - trace(Sigma * S_gradient[0]) / (h - sigma),
                              //
                              trace(transpose(S) * S_gradient[1]) / S_norm2                   //
                                + trace(transpose(Sigma) * Sigma_gradient[1]) / Sigma_norm2   //
                                - trace(Sigma * S_gradient[1]) / (h - sigma)};

      const TinyVector<Dimension> f_jr_gradient = f_jr * g;

      TinyMatrix<Dimension> f_jr_hessian = zero;
      for (size_t i = 0; i < Dimension; ++i) {
        for (size_t j = 0; j < Dimension; ++j) {
          f_jr_hessian(i, j) =                                                                                       //
            (trace(transpose(S_gradient[j]) * S_gradient[i]) / S_norm2                                               //
             - 2 * trace(transpose(S) * S_gradient[j]) * trace(transpose(S) * S_gradient[i]) / (S_norm2 * S_norm2)   //
                                                                                                                     //
             + trace(transpose(Sigma_gradient[j]) * Sigma_gradient[i]) / Sigma_norm2   // + 0
             - 2 * trace(transpose(Sigma) * Sigma_gradient[j]) * trace(transpose(Sigma) * Sigma_gradient[i]) /
                 (Sigma_norm2 * Sigma_norm2)   //

             - trace(Sigma_gradient[j] * S_gradient[i]) / (h - sigma)                     //
             + sigma / (std::pow(h - sigma, 3)) * sigma_gradient[i] * sigma_gradient[j]   //
             + g[j] * g[i]) *
            f_jr;
        }
      }

      f += f_jr;
      f_gradient += f_jr_gradient;
      f_hessian += f_jr_hessian;
    };

    auto escobar = escobarSSigma;

    f          = 0;
    f_gradient = zero;
    f_hessian  = zero;

    for (size_t i_cell = 0; i_cell < S_list.size(); ++i_cell) {
      const double sigma            = sigma_list[i_cell];
      const TinyMatrix<Dimension> S = S_list[i_cell];

      const TinyMatrix<Dimension> Sigma = [&S] {
        TinyMatrix<Dimension> transposed_comS;
        for (size_t i = 0; i < Dimension; ++i) {
          for (size_t j = 0; j < Dimension; ++j) {
            transposed_comS(i, j) = cofactor(S, j, i);
          }
        }
        return transposed_comS;
      }();

      TinyVector<Dimension> sigma_gradient{trace(Sigma * S_gradient[0]),   //
                                           trace(Sigma * S_gradient[1])};
      const std::array<TinyMatrix<Dimension>, Dimension>                           //
        Sigma_gradient{TinyMatrix<Dimension>{+0, +inv_sin_alpha - inv_tan_alpha,   //
                                             +0, -1},
                       TinyMatrix<Dimension>{-inv_sin_alpha + inv_tan_alpha, +0,   //
                                             +1, +0}};

      escobar(sigma, S, sigma_gradient, S_gradient, Sigma_gradient, f, f_gradient, f_hessian);
    }

    if constexpr (multi_connection) {
      for (size_t i_cell = 0; i_cell < cell_list.size(); ++i_cell) {
        const CellId cell_id       = cell_list[i_cell];
        auto cell_node_list        = cell_to_node_matrix[cell_list[i_cell]];
        const size_t cell_nb_nodes = cell_node_list.size();

        if (cell_nb_nodes <= 3) {
          continue;
        }

        const size_t i_cell_node = node_number_in_cell[i_cell];

        const double mr_alpha = 2 * std::acos(-1) / ((cell_nb_nodes - 2) * (cell_list.size()));

        const double mr_sin_alpha = std::sin(mr_alpha);
        const double mr_cos_alpha = std::cos(mr_alpha);
        // const double mr_inv_sin_alpha = 1. / mr_sin_alpha;
        // const double mr_inv_tan_alpha = 1. / std::tan(mr_alpha);

        const TinyMatrix<Dimension> mr_W{1, mr_cos_alpha,   //
                                         0, mr_sin_alpha};

        const TinyMatrix mr_inv_W = inverse(mr_W);

        std::array<TinyMatrix<Dimension>, Dimension>                                      //
          mr_S_gradient = {TinyMatrix<Dimension>{-1, (mr_cos_alpha - 1) / mr_sin_alpha,   //
                                                 +0, +0},
                           TinyMatrix<Dimension>{+0, +0,   //
                                                 -1, (mr_cos_alpha - 1) / mr_sin_alpha}};

        {   // first simplex
          const TinyVector a = old_xr[cell_node_list[(i_cell_node + 1) % cell_nb_nodes]];
          const TinyVector b = old_xr[cell_node_list[(i_cell_node + 2) % cell_nb_nodes]];

          const TinyMatrix<Dimension> mr_A{a[0] - x[0], b[0] - x[0],   //
                                           a[1] - x[1], b[1] - x[1]};

          const TinyMatrix mr_S = mr_A * mr_inv_W;
          const double mr_sigma = det(mr_S);

          const TinyMatrix<Dimension> mr_Sigma = [&mr_S] {
            TinyMatrix<Dimension> transposed_comS;
            for (size_t i = 0; i < Dimension; ++i) {
              for (size_t j = 0; j < Dimension; ++j) {
                transposed_comS(i, j) = cofactor(mr_S, j, i);
              }
            }
            return transposed_comS;
          }();

          TinyVector<Dimension> mr_sigma_gradient{trace(mr_Sigma * mr_S_gradient[0]),   //
                                                  trace(mr_Sigma * mr_S_gradient[1])};

          const std::array<TinyMatrix<Dimension>, Dimension>                                 //
            mr_Sigma_gradient{TinyMatrix<Dimension>{+0, (1 - mr_cos_alpha) / mr_sin_alpha,   //
                                                    +0, -1},
                              TinyMatrix<Dimension>{(mr_cos_alpha - 1) / mr_sin_alpha, +0,   //
                                                    +1, +0}};

          escobar(mr_sigma, mr_S, mr_sigma_gradient, mr_S_gradient, mr_Sigma_gradient, f, f_gradient, f_hessian);
        }

        {   // second simplex
          const TinyVector a = old_xr[cell_node_list[(i_cell_node + cell_nb_nodes - 2) % cell_nb_nodes]];
          const TinyVector b = old_xr[cell_node_list[(i_cell_node + cell_nb_nodes - 1) % cell_nb_nodes]];

          const TinyMatrix<Dimension> mr_A{a[0] - x[0], b[0] - x[0],   //
                                           a[1] - x[1], b[1] - x[1]};

          const TinyMatrix mr_S = mr_A * mr_inv_W;
          const double mr_sigma = det(mr_S);

          const TinyMatrix<Dimension> mr_Sigma = [&mr_S] {
            TinyMatrix<Dimension> transposed_comS;
            for (size_t i = 0; i < Dimension; ++i) {
              for (size_t j = 0; j < Dimension; ++j) {
                transposed_comS(i, j) = cofactor(mr_S, j, i);
              }
            }
            return transposed_comS;
          }();

          TinyVector<Dimension> mr_sigma_gradient{trace(mr_Sigma * mr_S_gradient[0]),   //
                                                  trace(mr_Sigma * mr_S_gradient[1])};

          const std::array<TinyMatrix<Dimension>, Dimension>                                 //
            mr_Sigma_gradient{TinyMatrix<Dimension>{+0, (1 - mr_cos_alpha) / mr_sin_alpha,   //
                                                    +0, -1},
                              TinyMatrix<Dimension>{(mr_cos_alpha - 1) / mr_sin_alpha, +0,   //
                                                    +1, +0}};

          escobar(mr_sigma, mr_S, mr_sigma_gradient, mr_S_gradient, mr_Sigma_gradient, f, f_gradient, f_hessian);
        }
      }
    }
  }

  NodeValue<const double>
  _getNodeQuality(const NodeValue<const TinyVector<Dimension>>& xr) const
  {
    if constexpr (Dimension == 2) {
      NodeValue<double> node_quality{m_given_mesh.connectivity()};

      auto node_to_cell_matrix        = m_given_mesh.connectivity().nodeToCellMatrix();
      auto cell_to_node_matrix        = m_given_mesh.connectivity().cellToNodeMatrix();
      auto node_number_in_their_cells = m_given_mesh.connectivity().nodeLocalNumbersInTheirCells();

      constexpr bool q_sin = true;

      for (NodeId node_id = 0; node_id < m_given_mesh.numberOfNodes(); ++node_id) {
        if (not m_is_fixed_node[node_id]) {
          auto node_cells  = node_to_cell_matrix[node_id];
          auto node_number = node_number_in_their_cells[node_id];

          if constexpr (q_sin) {   // q_area
            double min_sin     = std::numeric_limits<double>::max();
            double max_abs_sin = -std::numeric_limits<double>::max();

            for (size_t i_cell = 0; i_cell < node_cells.size(); ++i_cell) {
              const CellId cell_id = node_cells[i_cell];

              const auto cell_nodes = cell_to_node_matrix[cell_id];

              const size_t i_A = node_number[i_cell];
              const size_t i_B = (i_A + 1) % cell_nodes.size();
              const size_t i_C = (i_A + cell_nodes.size() - 1) % cell_nodes.size();
              const size_t i_D = (i_A + 2) % cell_nodes.size();
              const size_t i_E = (i_A + cell_nodes.size() - 2) % cell_nodes.size();

              const Rd A = xr[cell_nodes[i_A]];
              const Rd B = xr[cell_nodes[i_B]];
              const Rd C = xr[cell_nodes[i_C]];
              const Rd D = xr[cell_nodes[i_D]];
              const Rd E = xr[cell_nodes[i_E]];

              const Rd AB = B - A;
              const Rd AC = C - A;

              const Rd CA = A - C;
              const Rd CE = E - C;

              const Rd BD = D - B;
              const Rd BA = A - B;

              const double l_AB = l2Norm(AB);
              const double l_AC = l2Norm(AC);

              const double l_CA = l2Norm(CA);
              const double l_CE = l2Norm(CE);

              const double l_BD = l2Norm(BD);
              const double l_BA = l2Norm(BA);

              if ((l_AB == 0) or (l_AC == 0) or (l_CA == 0) or (l_CE == 0) or (l_BD == 0) or (l_BA == 0)) {
                min_sin     = -1;
                max_abs_sin = 1;
                break;
              }

              const Rd u_AB = 1. / l_AB * AB;
              const Rd u_AC = 1. / l_AC * AC;
              const Rd u_CA = 1. / l_CA * CA;
              const Rd u_CE = 1. / l_CE * CE;
              const Rd u_BD = 1. / l_BD * BD;
              const Rd u_BA = 1. / l_BA * BA;

              const double sin_CAB = det(TinyMatrix{u_AB, u_AC});
              const double sin_ECA = det(TinyMatrix{u_CA, u_CE});
              const double sin_ABD = det(TinyMatrix{u_BD, u_BA});

              const double abs_sin_CAB = std::abs(sin_CAB);
              const double abs_sin_ECA = std::abs(sin_ECA);
              const double abs_sin_ABD = std::abs(sin_ABD);

              if (sin_CAB < min_sin) {
                min_sin = sin_CAB;
              }

              if (sin_ECA < min_sin) {
                min_sin = sin_ECA;
              }

              if (sin_ABD < min_sin) {
                min_sin = sin_ABD;
              }

              if (abs_sin_CAB > max_abs_sin) {
                max_abs_sin = abs_sin_CAB;
              }

              if (abs_sin_ECA > max_abs_sin) {
                max_abs_sin = abs_sin_ECA;
              }

              if (abs_sin_ABD > max_abs_sin) {
                max_abs_sin = abs_sin_ABD;
              }
            }

            node_quality[node_id] = min_sin / max_abs_sin;
          } else {
            double min_area     = std::numeric_limits<double>::max();
            double max_abs_area = -std::numeric_limits<double>::max();

            for (size_t i_cell = 0; i_cell < node_cells.size(); ++i_cell) {
              const CellId cell_id = node_cells[i_cell];

              const auto cell_nodes = cell_to_node_matrix[cell_id];

              const size_t i_A = node_number[i_cell];
              const size_t i_B = (i_A + 1) % cell_nodes.size();
              const size_t i_C = (i_A + cell_nodes.size() - 1) % cell_nodes.size();
              const size_t i_D = (i_A + 2) % cell_nodes.size();
              const size_t i_E = (i_A + cell_nodes.size() - 2) % cell_nodes.size();

              const Rd A = xr[cell_nodes[i_A]];
              const Rd B = xr[cell_nodes[i_B]];
              const Rd C = xr[cell_nodes[i_C]];
              const Rd D = xr[cell_nodes[i_D]];
              const Rd E = xr[cell_nodes[i_E]];

              const Rd AB = B - A;
              const Rd AC = C - A;

              const Rd CA = A - C;
              const Rd CE = E - C;

              const Rd BD = D - B;
              const Rd BA = A - B;

              const double area_CAB = det(TinyMatrix{AB, AC});
              const double area_ECA = det(TinyMatrix{CA, CE});
              const double area_ABD = det(TinyMatrix{BD, BA});

              const double abs_area_CAB = std::abs(area_CAB);
              const double abs_area_ECA = std::abs(area_ECA);
              const double abs_area_ABD = std::abs(area_ABD);

              if (area_CAB < min_area) {
                min_area = area_CAB;
              }

              if (area_ECA < min_area) {
                min_area = area_ECA;
              }

              if (area_ABD < min_area) {
                min_area = area_ABD;
              }

              if (abs_area_CAB > max_abs_area) {
                max_abs_area = abs_area_CAB;
              }

              if (abs_area_ECA > max_abs_area) {
                max_abs_area = abs_area_ECA;
              }

              if (abs_area_ABD > max_abs_area) {
                max_abs_area = abs_area_ABD;
              }
            }

            node_quality[node_id] = min_area / max_abs_area;
          }
        } else {
          node_quality[node_id] = 1;
        }
      }
      return node_quality;

    } else {
      throw NotImplementedError("dimension != 2");
    }
  }

 public:
  std::shared_ptr<const ItemValueVariant>
  getQuality() const
  {
    return std::make_shared<ItemValueVariant>(this->_getNodeQuality(m_given_mesh.xr()));
  }

  void
  _getNewPositions(const NodeValue<const Rd>& old_xr, NodeValue<Rd>& new_xr) const
  {
    constexpr bool use_newton = false;

    const ConnectivityType& connectivity = m_given_mesh.connectivity();

    auto is_boundary_node = connectivity.isBoundaryNode();
    auto node_is_owned    = connectivity.nodeIsOwned();

    if constexpr (Dimension == 2) {
      auto cell_to_node_matrix        = connectivity.cellToNodeMatrix();
      auto node_to_cell_matrix        = connectivity.nodeToCellMatrix();
      auto node_to_edge_matrix        = connectivity.nodeToEdgeMatrix();
      auto edge_to_node_matrix        = connectivity.edgeToNodeMatrix();
      auto node_number_in_their_cells = connectivity.nodeLocalNumbersInTheirCells();

      NodeValue<double> quality{connectivity};

      quality.fill(2);

      auto smooth = [this, &old_xr](const NodeId node_id, TinyVector<Dimension>& x) -> double {
        double delta = 0;

        double final_f = 0;

        TinyMatrix<Dimension> H = identity;

        const ConnectivityType& connectivity = m_given_mesh.connectivity();

        for (size_t i_iter = 0; i_iter < 20; ++i_iter) {
          double sigma_min = std::numeric_limits<double>::max();
          delta            = sigma_min;

          double f                         = 0;
          TinyVector<Dimension> f_gradient = zero;
          TinyMatrix<Dimension> f_hessian  = zero;

          auto node_to_edge_matrix = connectivity.nodeToEdgeMatrix();
          auto edge_to_node_matrix = connectivity.edgeToNodeMatrix();

          get_f_grad_hess(old_xr, sigma_min, delta, node_id, x, f, f_gradient, f_hessian);

          const double max_edge_lenght = [=] {
            auto node_edge_list = node_to_edge_matrix[node_id];
            double max_lenght   = 0;
            for (size_t i_edge = 0; i_edge < node_edge_list.size(); ++i_edge) {
              const EdgeId edge_id            = node_edge_list[i_edge];
              auto edge_node_list             = edge_to_node_matrix[edge_id];
              const TinyVector<Dimension>& x0 = old_xr[edge_node_list[0]];
              const TinyVector<Dimension>& x1 = old_xr[edge_node_list[1]];
              const double lenght             = l2Norm(x0 - x1);
              if (lenght > max_lenght) {
                max_lenght = lenght;
              }
            }
            return max_lenght;
          }();

          if constexpr (use_newton) {
            const TinyVector delta_x_candidate = -inverse(f_hessian) * f_gradient;

            const double delta_x_norm = l2Norm(delta_x_candidate);

            TinyVector<Dimension> old_x = x;

            double factor = 1;
            if (delta_x_norm > 0.5 * max_edge_lenght) {
              factor *= (0.5 * max_edge_lenght / delta_x_norm);
            }

            x += factor * delta_x_candidate;

            if (delta_x_norm < 1E-8 * max_edge_lenght) {
              break;
            }
          } else {
            constexpr double c1 = 1e-4;
            constexpr double c2 = 0.7;

            auto zoom = [&](const TinyVector<Dimension>& x, const TinyVector<Dimension>& n, const double& f0,
                            const TinyVector<Dimension>& f_gradient0, const double& df0_dn, double alpha_low,
                            double alpha_high) -> double {
              double f_low                         = 0;
              TinyVector<Dimension> f_gradient_low = zero;
              TinyMatrix<Dimension> f_hessian_low  = zero;

              double init_delta = alpha_high - alpha_low;

              get_f_grad_hess(old_xr, sigma_min, delta, node_id, x - alpha_low * f_gradient0, f_low, f_gradient_low,
                              f_hessian_low);

              double f_high                         = 0;
              TinyVector<Dimension> f_gradient_high = zero;
              TinyMatrix<Dimension> f_hessian_high  = zero;

              get_f_grad_hess(old_xr, sigma_min, delta, node_id, x - alpha_high * f_gradient0, f_high, f_gradient_high,
                              f_hessian_high);

              // std::cout << "-- f_gradient     =" << f_gradient0 << " n=" << n << " df0_n=" << df0_dn << '\n';
              // std::cout << "   f_gradient_low =" << f_gradient_low << " n=" << n   //
              //           << " df_low_n=" << dot(f_gradient_low, n) << '\n';
              // std::cout << "   f_gradient_high=" << f_gradient_high << " n=" << n
              //           << " df_high_n=" << dot(f_gradient_high, n) << '\n';

              // std::cout << "starting box [" << alpha_low << ", " << alpha_high << "] --(f)--> [" << f_low << ", "
              //           << f_high << "]\n";

              for (size_t i = 1; i <= 200; ++i) {
                // std::cout << " box: " << alpha_low << " : " << alpha_high << '\n';
                if (alpha_high - alpha_low <= 1E-5 * init_delta) {
                  return alpha_low;
                }
                double alpha_i = 0.5 * (alpha_low + alpha_high);
                // std::cout << i << ": alpha_i = " << alpha_i << "[alpha_low : alpha_max] = " << alpha_low << " : "
                //           << alpha_high << "diff = " << alpha_high - alpha_low << '\n';

                double f_i                         = 0;
                TinyVector<Dimension> f_gradient_i = zero;
                TinyMatrix<Dimension> f_hessian_i  = zero;

                get_f_grad_hess(old_xr, sigma_min, delta, node_id, x - alpha_i * f_gradient0, f_i, f_gradient_i,
                                f_hessian_i);

                // std::cout << "   f_gradient_low =" << f_gradient_low << " n=" << n   //
                //           << " df_low_n=" << dot(f_gradient_low, n) << '\n';

                // std::cout << "f0 = " << f0 << " f_i = " << f_i << '\n';

                if ((f_i > f0 + c1 * alpha_i * df0_dn) or (f_i >= f_low)) {
                  alpha_high = alpha_i;
                  f_high     = f_i;
                } else {
                  double dfi_dn = dot(f_gradient_i, n);
                  if (std::abs(dfi_dn) <= -c2 * df0_dn) {
                    return alpha_i;
                  }
                  if (dfi_dn * (alpha_high - alpha_low) >= 0) {
                    alpha_high = alpha_low;
                    f_high     = f_low;
                  }

                  alpha_low = alpha_i;
                  f_low     = f_i;
                }
              }
              throw UnexpectedError("zoom did not converge!");
            };

            auto line_search = [&](const double& f0, const TinyVector<Dimension>& f_gradient0,
                                   const double alpha_max) -> double {
              double alpha_i_1 = 0;
              double alpha_i   = alpha_max;

              const TinyVector<Dimension> n = -1 / l2Norm(f_gradient0) * f_gradient0;

              const double df0_dn = dot(f_gradient0, n);

              double f_i_1 = f0;

              double f_i                         = 0;
              TinyVector<Dimension> f_gradient_i = zero;
              TinyMatrix<Dimension> f_hessian_i  = zero;

              for (size_t i = 1; i <= 100; ++i) {
                get_f_grad_hess(old_xr, sigma_min, delta, node_id, x - alpha_i * f_gradient0, f_i, f_gradient_i,
                                f_hessian_i);

                const double dfi_dn = dot(f_gradient_i, n);

                // std::cout << "=== i=" << i << " alpha_i=" << alpha_i << " alpha_i_1=" << alpha_i_1 << " f0=" <<
                // f0
                //           << " f_i=" << f_i << " f_i_1=" << f_i_1 << '\n';
                if ((f_i > f0 + c1 * alpha_i * df0_dn) or ((f_i >= f_i_1) and (i > 1))) {
                  // std::cout << "case 1: zoom[" << alpha_i << "," << alpha_i_1 << "] f_i=" << f_i << " f0=" << f0
                  //           << '\n';
                  return zoom(x, n, f0, f_gradient0, df0_dn, alpha_i_1, alpha_i);
                } else if (std::abs(dfi_dn) <= -c2 * df0_dn) {
                  // std::cout << "case 2: return alpha_i=" << alpha_i << '\n';
                  return alpha_i;
                } else if (dfi_dn >= 0) {
                  // std::cout << "case 3: zoom[" << alpha_i << "," << alpha_i_1 << "] f_i=" << f_i << " f0=" << f0
                  //           << '\n';
                  return zoom(x, n, f0, f_gradient0, df0_dn, alpha_i, alpha_i_1);
                } else {
                  alpha_i_1 = alpha_i;
                  f_i_1     = f_i;
                  alpha_i   = 0.5 * (alpha_i + alpha_max);
                }
              }
              throw UnexpectedError("line search failed");
            };

            if (l2Norm(f_gradient) < 1e-12 * f) {
              continue;
            }

            const double alpha_max = std::min(0.5 * max_edge_lenght / l2Norm(f_gradient), 1.);
            //            std::cout << rang::fgB::yellow << "i_iter=" << i_iter << rang::fg::reset << '\n';
            double alpha = line_search(f, f_gradient, alpha_max);

            const TinyVector delta_x = -alpha * H * f_gradient;

            double new_f                     = 0;
            TinyVector<Dimension> new_f_grad = zero;
            get_f_grad_hess(old_xr, sigma_min, delta, node_id, x + delta_x, new_f, new_f_grad, f_hessian);

            x += delta_x;

            // std::cout << "x -> " << x << '\n';
            // std::cout << rang::fgB::red << "f(x)=" << f << " f(x+dx)=" << new_f << " delta = " << delta_x
            //           << rang::fg::reset << '\n';
            // std::cout << rang::fgB::magenta << "l2Norm(delta_x) = " << l2Norm(delta_x) << rang::fg::reset <<
            // '\n';

            if (l2Norm(delta_x) < 1E-8 * max_edge_lenght) {
              break;
            } else {
              if (((i_iter + 1) % 5) == 0) {
                H = identity;
              } else {
                TinyVector y = new_f_grad - f_gradient;
                // std::cout << "l2Norm(y) = " << l2Norm(y) << '\n';
                const double rho = 1 / dot(delta_x, y);

                constexpr TinyMatrix<Dimension> I = identity;

                H = (I - rho * tensorProduct(delta_x, y)) * H * (I - rho * tensorProduct(y, delta_x)) +
                    rho * tensorProduct(delta_x, delta_x);
              }
            }
          }

          if (l2Norm(f_gradient) < 1E-12) {
            break;
          }

          final_f = f;
        }
        //        std::exit(0);

        return final_f;
      };

      parallel_for(
        connectivity.numberOfNodes(), PUGS_LAMBDA(const NodeId node_id) {
          if (node_is_owned[node_id] and m_is_smoothed_node[node_id] and not is_boundary_node[node_id]) {
            quality[node_id] = smooth(node_id, new_xr[node_id]);
            new_xr[node_id]  = 0.5 * (new_xr[node_id] + old_xr[node_id]);
          }
        });
      std::cout << "nb smoothed = " << [&] {
        size_t count = 0;
        for (NodeId node_id = 0; node_id < connectivity.numberOfNodes(); ++node_id) {
          count += (node_is_owned[node_id] and m_is_smoothed_node[node_id] and not is_boundary_node[node_id]);
        }
        return count;
      }() << '\n';
    } else {
      throw NotImplementedError("Dimension != 2");
    }
    this->_applyBC(old_xr, new_xr);
  }

  std::shared_ptr<const IMesh>
  _smoothedMesh()
  {
    NodeValue<const Rd> xr = m_given_mesh.xr();

    size_t i_convexify  = 1;
    bool need_convexify = false;

    std::cout << "== calling smoothMesh ==\n";

    NodeValue<const double> old_node_quality;
    do {
      std::cout << "- convexfication step " << i_convexify++ << '\n';
      NodeValue<Rd> new_xr = copy(xr);

      auto node_quality = this->_getNodeQuality(xr);

      need_convexify = min(node_quality) <= 0;

      if (need_convexify) {
        m_is_smoothed_node = _getSmoothedNode(node_quality, m_is_fixed_node);

        if (old_node_quality.isBuilt()) {
          NodeValue<bool> is_smoothed_first{m_given_mesh.connectivity()};
          is_smoothed_first.fill(false);
          bool has_new_smoothed_node = false;
          for (NodeId node_id = 0; node_id < m_given_mesh.numberOfNodes(); ++node_id) {
            if ((m_is_smoothed_node[node_id]) and (old_node_quality[node_id] >= 0)) {
              is_smoothed_first[node_id] = true;
              has_new_smoothed_node      = true;
            }
          }

          if (has_new_smoothed_node) {
            std::cout << "has_new_smoothed_node = " << has_new_smoothed_node << '\n';
            m_is_smoothed_node = is_smoothed_first;
          }
        }

        this->_getNewPositions(xr, new_xr);
        xr = new_xr;
      }

      old_node_quality = node_quality;
      std::cout << "need_convexify = " << std::boolalpha << need_convexify << '\n';
    } while ((need_convexify) and (i_convexify < 50));
    std::cout << " --> finished\n";

    return std::make_shared<MeshType>(m_given_mesh.shared_connectivity(), xr);
  }

  std::shared_ptr<const IMesh>
  getSmoothedMesh()
  {
    NodeValue<bool> is_smoothed_node(m_given_mesh.connectivity());
    parallel_for(
      m_given_mesh.numberOfNodes(),
      PUGS_LAMBDA(const NodeId node_id) { is_smoothed_node[node_id] = not m_is_fixed_node[node_id]; });

    m_is_smoothed_node = is_smoothed_node;

    return _smoothedMesh();
  }

  std::shared_ptr<const IMesh>
  getSmoothedMesh(const FunctionSymbolId& function_symbol_id)
  {
    NodeValue<const Rd> given_xr = m_given_mesh.xr();

    NodeValue<const bool> is_displaced =
      InterpolateItemValue<bool(const Rd)>::interpolate(function_symbol_id, given_xr);

    NodeValue<bool> is_fixed_node(m_given_mesh.connectivity());
    parallel_for(
      m_given_mesh.numberOfNodes(), PUGS_LAMBDA(const NodeId node_id) {
        is_fixed_node[node_id] = m_is_fixed_node[node_id] or not is_displaced[node_id];
      });

    synchronize(is_fixed_node);

    m_is_fixed_node = is_fixed_node;

    return _smoothedMesh();
  }

  std::shared_ptr<const IMesh>
  getSmoothedMesh(const std::vector<std::shared_ptr<const IZoneDescriptor>>& zone_descriptor_list)
  {
    NodeValue<const Rd> given_xr = m_given_mesh.xr();

    auto node_to_cell_matrix = m_given_mesh.connectivity().nodeToCellMatrix();

    NodeValue<bool> is_displaced{m_given_mesh.connectivity()};
    is_displaced.fill(false);

    for (size_t i_zone = 0; i_zone < zone_descriptor_list.size(); ++i_zone) {
      MeshCellZone<Dimension> cell_zone = getMeshCellZone(m_given_mesh, *zone_descriptor_list[i_zone]);
      const auto cell_list              = cell_zone.cellList();
      CellValue<bool> is_zone_cell{m_given_mesh.connectivity()};
      is_zone_cell.fill(false);
      parallel_for(
        cell_list.size(), PUGS_LAMBDA(const size_t i_cell) { is_zone_cell[cell_list[i_cell]] = true; });
      parallel_for(
        m_given_mesh.numberOfNodes(), PUGS_LAMBDA(const NodeId node_id) {
          auto node_cell_list = node_to_cell_matrix[node_id];
          bool displace       = true;
          for (size_t i_node_cell = 0; i_node_cell < node_cell_list.size(); ++i_node_cell) {
            const CellId cell_id = node_cell_list[i_node_cell];
            displace &= is_zone_cell[cell_id];
          }
          if (displace) {
            is_displaced[node_id] = true;
          }
        });
    }
    synchronize(is_displaced);

    NodeValue<bool> is_fixed_node(m_given_mesh.connectivity());
    parallel_for(
      m_given_mesh.numberOfNodes(), PUGS_LAMBDA(const NodeId node_id) {
        is_fixed_node[node_id] = m_is_fixed_node[node_id] or not is_displaced[node_id];
      });

    synchronize(is_fixed_node);

    m_is_fixed_node = is_fixed_node;

    return _smoothedMesh();
  }

  std::shared_ptr<const IMesh>
  getSmoothedMesh(const std::vector<std::shared_ptr<const DiscreteFunctionVariant>>& discrete_function_variant_list)
  {
    NodeValue<const Rd> given_xr = m_given_mesh.xr();

    auto node_to_cell_matrix = m_given_mesh.connectivity().nodeToCellMatrix();

    CellValue<bool> is_zone_cell{m_given_mesh.connectivity()};
    is_zone_cell.fill(false);

    for (size_t i_zone = 0; i_zone < discrete_function_variant_list.size(); ++i_zone) {
      auto characteristic_function =
        discrete_function_variant_list[i_zone]->get<DiscreteFunctionP0<Dimension, const double>>();
      parallel_for(
        m_given_mesh.numberOfCells(), PUGS_LAMBDA(const CellId cell_id) {
          if (characteristic_function[cell_id] != 0) {
            is_zone_cell[cell_id] = true;
          }
        });
    }

    CellValue<const int> non_convex_cell_tag = this->_getNonConvexCellTag(given_xr);

    const bool has_non_convex_cell = max(non_convex_cell_tag) > 0;

    if (not has_non_convex_cell) {
      return m_p_given_mesh;
    }

    NodeValue<bool> is_displaced{m_given_mesh.connectivity()};
    is_displaced.fill(false);

    for (size_t i_zone = 0; i_zone < discrete_function_variant_list.size(); ++i_zone) {
      auto characteristic_function =
        discrete_function_variant_list[i_zone]->get<DiscreteFunctionP0<Dimension, const double>>();

      parallel_for(
        m_given_mesh.numberOfNodes(), PUGS_LAMBDA(const NodeId node_id) {
          auto node_cell_list = node_to_cell_matrix[node_id];
          bool displace       = true;
          for (size_t i_node_cell = 0; i_node_cell < node_cell_list.size(); ++i_node_cell) {
            const CellId cell_id = node_cell_list[i_node_cell];
            displace &= (characteristic_function[cell_id] != 0);
          }
          if (displace) {
            is_displaced[node_id] = true;
          }
        });
    }
    synchronize(is_displaced);

    NodeValue<bool> is_fixed_node(m_given_mesh.connectivity());
    parallel_for(
      m_given_mesh.numberOfNodes(), PUGS_LAMBDA(const NodeId node_id) {
        is_fixed_node[node_id] = m_is_fixed_node[node_id] or not is_displaced[node_id];
      });

    synchronize(is_fixed_node);

    m_is_fixed_node = is_fixed_node;

    m_is_smoothed_node = _getSmoothedNode(non_convex_cell_tag, m_is_fixed_node);

    return _smoothedMesh();
  }

  MeshSmootherEscobar(const MeshSmootherEscobar&) = delete;
  MeshSmootherEscobar(MeshSmootherEscobar&&)      = delete;

  MeshSmootherEscobar(const MeshType& given_mesh,
                      const std::shared_ptr<const IMesh>& p_mesh,
                      const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>& bc_descriptor_list)
    : m_given_mesh{given_mesh},
      m_p_given_mesh{p_mesh},
      m_boundary_condition_list{this->_getBCList(given_mesh, bc_descriptor_list)},
      m_is_fixed_node{this->_getFixedNodes()}
  {}

  ~MeshSmootherEscobar() = default;
};

template <size_t Dimension>
class MeshSmootherEscobarHandler::MeshSmootherEscobar<Dimension>::AxisBoundaryCondition
{
 public:
  using Rd = TinyVector<Dimension, double>;

 private:
  const MeshLineNodeBoundary<Dimension> m_mesh_line_node_boundary;

 public:
  const Rd&
  direction() const
  {
    return m_mesh_line_node_boundary.direction();
  }

  const Array<const NodeId>&
  nodeList() const
  {
    return m_mesh_line_node_boundary.nodeList();
  }

  AxisBoundaryCondition(MeshLineNodeBoundary<Dimension>&& mesh_line_node_boundary)
    : m_mesh_line_node_boundary(mesh_line_node_boundary)
  {
    ;
  }

  ~AxisBoundaryCondition() = default;
};

template <>
class MeshSmootherEscobarHandler::MeshSmootherEscobar<1>::AxisBoundaryCondition
{
 public:
  AxisBoundaryCondition()  = default;
  ~AxisBoundaryCondition() = default;
};

template <size_t Dimension>
class MeshSmootherEscobarHandler::MeshSmootherEscobar<Dimension>::FixedBoundaryCondition
{
 private:
  const MeshNodeBoundary<Dimension> m_mesh_node_boundary;

 public:
  const Array<const NodeId>&
  nodeList() const
  {
    return m_mesh_node_boundary.nodeList();
  }

  FixedBoundaryCondition(MeshNodeBoundary<Dimension>&& mesh_node_boundary) : m_mesh_node_boundary{mesh_node_boundary} {}

  ~FixedBoundaryCondition() = default;
};

template <size_t Dimension>
class MeshSmootherEscobarHandler::MeshSmootherEscobar<Dimension>::SymmetryBoundaryCondition
{
 public:
  using Rd = TinyVector<Dimension, double>;

 private:
  const MeshFlatNodeBoundary<Dimension> m_mesh_flat_node_boundary;

 public:
  const Rd&
  outgoingNormal() const
  {
    return m_mesh_flat_node_boundary.outgoingNormal();
  }

  const Array<const NodeId>&
  nodeList() const
  {
    return m_mesh_flat_node_boundary.nodeList();
  }

  SymmetryBoundaryCondition(MeshFlatNodeBoundary<Dimension>&& mesh_flat_node_boundary)
    : m_mesh_flat_node_boundary(mesh_flat_node_boundary)
  {
    ;
  }

  ~SymmetryBoundaryCondition() = default;
};

std::shared_ptr<const ItemValueVariant>
MeshSmootherEscobarHandler::getQuality(
  const std::shared_ptr<const IMesh>& mesh,
  const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>& bc_descriptor_list) const
{
  switch (mesh->dimension()) {
  case 1: {
    constexpr size_t Dimension = 1;
    using MeshType             = Mesh<Connectivity<Dimension>>;
    MeshSmootherEscobar smoother(dynamic_cast<const MeshType&>(*mesh), mesh, bc_descriptor_list);
    return smoother.getQuality();
  }
  case 2: {
    constexpr size_t Dimension = 2;
    using MeshType             = Mesh<Connectivity<Dimension>>;
    MeshSmootherEscobar smoother(dynamic_cast<const MeshType&>(*mesh), mesh, bc_descriptor_list);
    return smoother.getQuality();
  }
  case 3: {
    constexpr size_t Dimension = 3;
    using MeshType             = Mesh<Connectivity<Dimension>>;
    MeshSmootherEscobar smoother(dynamic_cast<const MeshType&>(*mesh), mesh, bc_descriptor_list);
    return smoother.getQuality();
  }
  default: {
    throw UnexpectedError("invalid mesh dimension");
  }
  }
}

std::shared_ptr<const IMesh>
MeshSmootherEscobarHandler::getSmoothedMesh(
  const std::shared_ptr<const IMesh>& mesh,
  const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>& bc_descriptor_list) const
{
  switch (mesh->dimension()) {
  case 1: {
    constexpr size_t Dimension = 1;
    using MeshType             = Mesh<Connectivity<Dimension>>;
    MeshSmootherEscobar smoother(dynamic_cast<const MeshType&>(*mesh), mesh, bc_descriptor_list);
    return smoother.getSmoothedMesh();
  }
  case 2: {
    constexpr size_t Dimension = 2;
    using MeshType             = Mesh<Connectivity<Dimension>>;
    MeshSmootherEscobar smoother(dynamic_cast<const MeshType&>(*mesh), mesh, bc_descriptor_list);
    return smoother.getSmoothedMesh();
  }
  case 3: {
    constexpr size_t Dimension = 3;
    using MeshType             = Mesh<Connectivity<Dimension>>;
    MeshSmootherEscobar smoother(dynamic_cast<const MeshType&>(*mesh), mesh, bc_descriptor_list);
    return smoother.getSmoothedMesh();
  }
  default: {
    throw UnexpectedError("invalid mesh dimension");
  }
  }
}

std::shared_ptr<const IMesh>
MeshSmootherEscobarHandler::getSmoothedMesh(
  const std::shared_ptr<const IMesh>& mesh,
  const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>& bc_descriptor_list,
  const FunctionSymbolId& function_symbol_id) const
{
  switch (mesh->dimension()) {
  case 1: {
    constexpr size_t Dimension = 1;
    using MeshType             = Mesh<Connectivity<Dimension>>;
    MeshSmootherEscobar smoother(dynamic_cast<const MeshType&>(*mesh), mesh, bc_descriptor_list);
    return smoother.getSmoothedMesh(function_symbol_id);
  }
  case 2: {
    constexpr size_t Dimension = 2;
    using MeshType             = Mesh<Connectivity<Dimension>>;
    MeshSmootherEscobar smoother(dynamic_cast<const MeshType&>(*mesh), mesh, bc_descriptor_list);
    return smoother.getSmoothedMesh(function_symbol_id);
  }
  case 3: {
    constexpr size_t Dimension = 3;
    using MeshType             = Mesh<Connectivity<Dimension>>;
    MeshSmootherEscobar smoother(dynamic_cast<const MeshType&>(*mesh), mesh, bc_descriptor_list);
    return smoother.getSmoothedMesh(function_symbol_id);
  }
  default: {
    throw UnexpectedError("invalid mesh dimension");
  }
  }
}

std::shared_ptr<const IMesh>
MeshSmootherEscobarHandler::getSmoothedMesh(
  const std::shared_ptr<const IMesh>& mesh,
  const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>& bc_descriptor_list,
  const std::vector<std::shared_ptr<const IZoneDescriptor>>& smoothing_zone_list) const
{
  switch (mesh->dimension()) {
  case 1: {
    constexpr size_t Dimension = 1;
    using MeshType             = Mesh<Connectivity<Dimension>>;
    MeshSmootherEscobar smoother(dynamic_cast<const MeshType&>(*mesh), mesh, bc_descriptor_list);
    return smoother.getSmoothedMesh(smoothing_zone_list);
  }
  case 2: {
    constexpr size_t Dimension = 2;
    using MeshType             = Mesh<Connectivity<Dimension>>;
    MeshSmootherEscobar smoother(dynamic_cast<const MeshType&>(*mesh), mesh, bc_descriptor_list);
    return smoother.getSmoothedMesh(smoothing_zone_list);
  }
  case 3: {
    constexpr size_t Dimension = 3;
    using MeshType             = Mesh<Connectivity<Dimension>>;
    MeshSmootherEscobar smoother(dynamic_cast<const MeshType&>(*mesh), mesh, bc_descriptor_list);
    return smoother.getSmoothedMesh(smoothing_zone_list);
  }
  default: {
    throw UnexpectedError("invalid mesh dimension");
  }
  }
}

std::shared_ptr<const IMesh>
MeshSmootherEscobarHandler::getSmoothedMesh(
  const std::shared_ptr<const IMesh>& mesh,
  const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>& bc_descriptor_list,
  const std::vector<std::shared_ptr<const DiscreteFunctionVariant>>& discrete_function_variant_list) const
{
  if (not hasSameMesh(discrete_function_variant_list)) {
    throw NormalError("discrete functions are not defined on the same mesh");
  }

  std::shared_ptr<const IMesh> common_mesh = getCommonMesh(discrete_function_variant_list);

  if (common_mesh != mesh) {
    throw NormalError("discrete functions are not defined on the smoothed mesh");
  }

  switch (mesh->dimension()) {
  case 1: {
    constexpr size_t Dimension = 1;
    using MeshType             = Mesh<Connectivity<Dimension>>;
    MeshSmootherEscobar smoother(dynamic_cast<const MeshType&>(*mesh), mesh, bc_descriptor_list);
    return smoother.getSmoothedMesh(discrete_function_variant_list);
  }
  case 2: {
    constexpr size_t Dimension = 2;
    using MeshType             = Mesh<Connectivity<Dimension>>;
    MeshSmootherEscobar smoother(dynamic_cast<const MeshType&>(*mesh), mesh, bc_descriptor_list);
    return smoother.getSmoothedMesh(discrete_function_variant_list);
  }
  case 3: {
    constexpr size_t Dimension = 3;
    using MeshType             = Mesh<Connectivity<Dimension>>;
    MeshSmootherEscobar smoother(dynamic_cast<const MeshType&>(*mesh), mesh, bc_descriptor_list);
    return smoother.getSmoothedMesh(discrete_function_variant_list);
  }
  default: {
    throw UnexpectedError("invalid mesh dimension");
  }
  }
}
