#include <catch2/catch_test_macros.hpp>
#include <catch2/matchers/catch_matchers_all.hpp>

#include <language/ast/ASTBuilder.hpp>
#include <language/ast/ASTModulesImporter.hpp>
#include <language/ast/ASTNodeDataTypeBuilder.hpp>
#include <language/ast/ASTNodeExpressionBuilder.hpp>
#include <language/ast/ASTNodeFunctionEvaluationExpressionBuilder.hpp>
#include <language/ast/ASTNodeFunctionExpressionBuilder.hpp>
#include <language/ast/ASTNodeTypeCleaner.hpp>
#include <language/ast/ASTSymbolTableBuilder.hpp>
#include <language/utils/PugsFunctionAdapter.hpp>
#include <language/utils/SymbolTable.hpp>

#include <MeshDataBaseForTests.hpp>
#include <mesh/Connectivity.hpp>
#include <mesh/Mesh.hpp>
#include <mesh/MeshData.hpp>
#include <mesh/MeshDataManager.hpp>

#include <language/utils/ItemValueVariantFunctionInterpoler.hpp>

#include <pegtl/string_input.hpp>

// clazy:excludeall=non-pod-global-static

TEST_CASE("ItemValueVariantFunctionInterpoler", "[scheme]")
{
  const bool stack_details = ASTNode::getStackDetails();
  ASTNode::setStackDetails(false);

  auto same_item_value = [](auto f, auto g) -> bool {
    using ItemIdType = typename decltype(f)::index_type;
    for (ItemIdType item_id = 0; item_id < f.numberOfItems(); ++item_id) {
      if (f[item_id] != g[item_id]) {
        return false;
      }
    }

    return true;
  };

  SECTION("1D")
  {
    constexpr size_t Dimension = 1;

    std::array mesh_list = MeshDataBaseForTests::get().all1DMeshes();

    for (const auto& named_mesh : mesh_list) {
      SECTION(named_mesh.name())
      {
        auto mesh_1d = named_mesh.mesh();

        auto xj = MeshDataManager::instance().getMeshData(*mesh_1d).xj();
        auto xr = mesh_1d->xr();

        std::string_view data = R"(
import math;
let B_scalar_non_linear_1d: R^1 -> B, x -> (exp(2 * x[0]) + 3 > 4);
let N_scalar_non_linear_1d: R^1 -> N, x -> floor(3 * x[0] * x[0] + 2);
let Z_scalar_non_linear_1d: R^1 -> Z, x -> floor(exp(2 * x[0]) - 1);
let R_scalar_non_linear_1d: R^1 -> R, x -> 2 * exp(x[0]) + 3;
let R1_non_linear_1d: R^1 -> R^1, x -> 2 * exp(x[0]);
let R2_non_linear_1d: R^1 -> R^2, x -> [2 * exp(x[0]), -3*x[0]];
let R3_non_linear_1d: R^1 -> R^3, x -> [2 * exp(x[0]) + 3, x[0] - 2, 3];
let R1x1_non_linear_1d: R^1 -> R^1x1, x -> (2 * exp(x[0]) * sin(x[0]) + 3);
let R2x2_non_linear_1d: R^1 -> R^2x2, x -> [[2 * exp(x[0]) * sin(x[0]) + 3, sin(x[0] - 2 * x[0])], [3, x[0] * x[0]]];
let R3x3_non_linear_1d: R^1 -> R^3x3, x -> [[2 * exp(x[0]) * sin(x[0]) + 3, sin(x[0] - 2 * x[0]), 3], [x[0] * x[0], -4*x[0], 2*x[0]+1], [3, -6*x[0], exp(x[0])]];
)";
        TAO_PEGTL_NAMESPACE::string_input input{data, "test.pgs"};

        auto ast = ASTBuilder::build(input);

        ASTModulesImporter{*ast};
        ASTNodeTypeCleaner<language::import_instruction>{*ast};

        ASTSymbolTableBuilder{*ast};
        ASTNodeDataTypeBuilder{*ast};

        ASTNodeTypeCleaner<language::var_declaration>{*ast};
        ASTNodeTypeCleaner<language::fct_declaration>{*ast};
        ASTNodeExpressionBuilder{*ast};

        std::shared_ptr<SymbolTable> symbol_table = ast->m_symbol_table;

        // ensure that variables are declared at this point
        TAO_PEGTL_NAMESPACE::position position{data.size(), 1, 1, "fixture"};

        SECTION("B_scalar_non_linear_1d")
        {
          auto [i_symbol, found] = symbol_table->find("B_scalar_non_linear_1d", position);
          REQUIRE(found);
          REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

          FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

          CellValue<bool> cell_value{mesh_1d->connectivity()};
          parallel_for(
            cell_value.numberOfItems(), PUGS_LAMBDA(const CellId cell_id) {
              const TinyVector<Dimension>& x = xj[cell_id];
              cell_value[cell_id]            = std::exp(2 * x[0]) + 3 > 4;
            });

          ItemValueVariantFunctionInterpoler interpoler(mesh_1d, ItemType::cell, function_symbol_id);
          std::shared_ptr item_data_variant = interpoler.interpolate();

          REQUIRE(same_item_value(cell_value, item_data_variant->get<CellValue<bool>>()));
        }

        SECTION("N_scalar_non_linear_1d")
        {
          auto [i_symbol, found] = symbol_table->find("N_scalar_non_linear_1d", position);
          REQUIRE(found);
          REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

          FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

          NodeValue<uint64_t> node_value{mesh_1d->connectivity()};
          parallel_for(
            node_value.numberOfItems(), PUGS_LAMBDA(const NodeId node_id) {
              const TinyVector<Dimension>& x = xr[node_id];
              node_value[node_id]            = std::floor(3 * x[0] * x[0] + 2);
            });

          ItemValueVariantFunctionInterpoler interpoler(mesh_1d, ItemType::node, function_symbol_id);
          std::shared_ptr item_data_variant = interpoler.interpolate();

          REQUIRE(same_item_value(node_value, item_data_variant->get<NodeValue<uint64_t>>()));
        }

        SECTION("Z_scalar_non_linear_1d")
        {
          auto [i_symbol, found] = symbol_table->find("Z_scalar_non_linear_1d", position);
          REQUIRE(found);
          REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

          FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

          CellValue<int64_t> cell_value{mesh_1d->connectivity()};
          parallel_for(
            cell_value.numberOfItems(), PUGS_LAMBDA(const CellId cell_id) {
              const TinyVector<Dimension>& x = xj[cell_id];
              cell_value[cell_id]            = std::floor(std::exp(2 * x[0]) - 1);
            });

          ItemValueVariantFunctionInterpoler interpoler(mesh_1d, ItemType::cell, function_symbol_id);
          std::shared_ptr item_value_variant = interpoler.interpolate();

          REQUIRE(same_item_value(cell_value, item_value_variant->get<CellValue<int64_t>>()));
        }

        SECTION("R_scalar_non_linear_1d")
        {
          auto [i_symbol, found] = symbol_table->find("R_scalar_non_linear_1d", position);
          REQUIRE(found);
          REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

          FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

          CellValue<double> cell_value{mesh_1d->connectivity()};
          parallel_for(
            cell_value.numberOfItems(), PUGS_LAMBDA(const CellId cell_id) {
              const TinyVector<Dimension>& x = xj[cell_id];
              cell_value[cell_id]            = 2 * std::exp(x[0]) + 3;
            });

          ItemValueVariantFunctionInterpoler interpoler(mesh_1d, ItemType::cell, function_symbol_id);
          std::shared_ptr item_value_variant = interpoler.interpolate();

          REQUIRE(same_item_value(cell_value, item_value_variant->get<CellValue<double>>()));
        }

        SECTION("R1_non_linear_1d")
        {
          using DataType = TinyVector<1>;

          auto [i_symbol, found] = symbol_table->find("R1_non_linear_1d", position);
          REQUIRE(found);
          REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

          FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

          CellValue<DataType> cell_value{mesh_1d->connectivity()};
          parallel_for(
            cell_value.numberOfItems(), PUGS_LAMBDA(const CellId cell_id) {
              const TinyVector<Dimension>& x = xj[cell_id];
              cell_value[cell_id]            = DataType{2 * std::exp(x[0])};
            });

          ItemValueVariantFunctionInterpoler interpoler(mesh_1d, ItemType::cell, function_symbol_id);
          std::shared_ptr item_value_variant = interpoler.interpolate();

          REQUIRE(same_item_value(cell_value, item_value_variant->get<CellValue<DataType>>()));
        }

        SECTION("R2_non_linear_1d")
        {
          using DataType = TinyVector<2>;

          auto [i_symbol, found] = symbol_table->find("R2_non_linear_1d", position);
          REQUIRE(found);
          REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

          FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

          CellValue<DataType> cell_value{mesh_1d->connectivity()};
          parallel_for(
            cell_value.numberOfItems(), PUGS_LAMBDA(const CellId cell_id) {
              const TinyVector<Dimension>& x = xj[cell_id];
              cell_value[cell_id]            = DataType{2 * std::exp(x[0]), -3 * x[0]};
            });

          ItemValueVariantFunctionInterpoler interpoler(mesh_1d, ItemType::cell, function_symbol_id);
          std::shared_ptr item_value_variant = interpoler.interpolate();

          REQUIRE(same_item_value(cell_value, item_value_variant->get<CellValue<DataType>>()));
        }

        SECTION("R3_non_linear_1d")
        {
          using DataType = TinyVector<3>;

          auto [i_symbol, found] = symbol_table->find("R3_non_linear_1d", position);
          REQUIRE(found);
          REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

          FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

          CellValue<DataType> cell_value{mesh_1d->connectivity()};
          parallel_for(
            cell_value.numberOfItems(), PUGS_LAMBDA(const CellId cell_id) {
              const TinyVector<Dimension>& x = xj[cell_id];
              cell_value[cell_id]            = DataType{2 * std::exp(x[0]) + 3, x[0] - 2, 3};
            });

          ItemValueVariantFunctionInterpoler interpoler(mesh_1d, ItemType::cell, function_symbol_id);
          std::shared_ptr item_value_variant = interpoler.interpolate();

          REQUIRE(same_item_value(cell_value, item_value_variant->get<CellValue<DataType>>()));
        }

        SECTION("R1x1_non_linear_1d")
        {
          using DataType = TinyMatrix<1>;

          auto [i_symbol, found] = symbol_table->find("R1x1_non_linear_1d", position);
          REQUIRE(found);
          REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

          FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

          CellValue<DataType> cell_value{mesh_1d->connectivity()};
          parallel_for(
            cell_value.numberOfItems(), PUGS_LAMBDA(const CellId cell_id) {
              const TinyVector<Dimension>& x = xj[cell_id];
              cell_value[cell_id]            = DataType{2 * std::exp(x[0]) * std::sin(x[0]) + 3};
            });

          ItemValueVariantFunctionInterpoler interpoler(mesh_1d, ItemType::cell, function_symbol_id);
          std::shared_ptr item_value_variant = interpoler.interpolate();

          REQUIRE(same_item_value(cell_value, item_value_variant->get<CellValue<DataType>>()));
        }

        SECTION("R2x2_non_linear_1d")
        {
          using DataType = TinyMatrix<2>;

          auto [i_symbol, found] = symbol_table->find("R2x2_non_linear_1d", position);
          REQUIRE(found);
          REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

          FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

          CellValue<DataType> cell_value{mesh_1d->connectivity()};
          parallel_for(
            cell_value.numberOfItems(), PUGS_LAMBDA(const CellId cell_id) {
              const TinyVector<Dimension>& x = xj[cell_id];
              cell_value[cell_id] =
                DataType{2 * std::exp(x[0]) * std::sin(x[0]) + 3, std::sin(x[0] - 2 * x[0]), 3, x[0] * x[0]};
            });

          ItemValueVariantFunctionInterpoler interpoler(mesh_1d, ItemType::cell, function_symbol_id);
          std::shared_ptr item_value_variant = interpoler.interpolate();

          REQUIRE(same_item_value(cell_value, item_value_variant->get<CellValue<DataType>>()));
        }

        SECTION("R3x3_non_linear_1d")
        {
          using DataType = TinyMatrix<3>;

          auto [i_symbol, found] = symbol_table->find("R3x3_non_linear_1d", position);
          REQUIRE(found);
          REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

          FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

          CellValue<DataType> cell_value{mesh_1d->connectivity()};
          parallel_for(
            cell_value.numberOfItems(), PUGS_LAMBDA(const CellId cell_id) {
              const TinyVector<Dimension>& x = xj[cell_id];
              cell_value[cell_id]            = DataType{2 * exp(x[0]) * std::sin(x[0]) + 3,
                                             std::sin(x[0] - 2 * x[0]),
                                             3,
                                             x[0] * x[0],
                                             -4 * x[0],
                                             2 * x[0] + 1,
                                             3,
                                             -6 * x[0],
                                             std::exp(x[0])};
            });

          ItemValueVariantFunctionInterpoler interpoler(mesh_1d, ItemType::cell, function_symbol_id);
          std::shared_ptr item_value_variant = interpoler.interpolate();

          REQUIRE(same_item_value(cell_value, item_value_variant->get<CellValue<DataType>>()));
        }
      }
    }
  }

  SECTION("2D")
  {
    constexpr size_t Dimension = 2;

    std::array mesh_list = MeshDataBaseForTests::get().all2DMeshes();

    for (const auto& named_mesh : mesh_list) {
      SECTION(named_mesh.name())
      {
        auto mesh_2d = named_mesh.mesh();

        auto xl = MeshDataManager::instance().getMeshData(*mesh_2d).xl();

        std::string_view data = R"(
import math;
let B_scalar_non_linear_2d: R^2 -> B, x -> (exp(2 * x[0])< 2*x[1]);
let R2_non_linear_2d: R^2 -> R^2, x -> [2 * exp(x[0]), -3*x[1]];
let R3x3_non_linear_2d: R^2 -> R^3x3, x -> [[2 * exp(x[0]) * sin(x[1]) + 3, sin(x[1] - 2 * x[0]), 3], [x[1] * x[0], -4*x[1], 2*x[0]+1], [3, -6*x[0], exp(x[1])]];
)";
        TAO_PEGTL_NAMESPACE::string_input input{data, "test.pgs"};

        auto ast = ASTBuilder::build(input);

        ASTModulesImporter{*ast};
        ASTNodeTypeCleaner<language::import_instruction>{*ast};

        ASTSymbolTableBuilder{*ast};
        ASTNodeDataTypeBuilder{*ast};

        ASTNodeTypeCleaner<language::var_declaration>{*ast};
        ASTNodeTypeCleaner<language::fct_declaration>{*ast};
        ASTNodeExpressionBuilder{*ast};

        std::shared_ptr<SymbolTable> symbol_table = ast->m_symbol_table;

        // ensure that variables are declared at this point
        TAO_PEGTL_NAMESPACE::position position{data.size(), 1, 1, "fixture"};

        SECTION("B_scalar_non_linear_2d")
        {
          using DataType = bool;

          auto [i_symbol, found] = symbol_table->find("B_scalar_non_linear_2d", position);
          REQUIRE(found);
          REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

          FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

          FaceValue<DataType> face_value{mesh_2d->connectivity()};
          parallel_for(
            face_value.numberOfItems(), PUGS_LAMBDA(const FaceId face_id) {
              const TinyVector<Dimension>& x = xl[face_id];
              face_value[face_id]            = std::exp(2 * x[0]) < 2 * x[1];
            });

          ItemValueVariantFunctionInterpoler interpoler(mesh_2d, ItemType::face, function_symbol_id);
          std::shared_ptr item_value_variant = interpoler.interpolate();

          REQUIRE(same_item_value(face_value, item_value_variant->get<FaceValue<DataType>>()));
        }

        SECTION("R2_non_linear_2d")
        {
          using DataType = TinyVector<2>;

          auto [i_symbol, found] = symbol_table->find("R2_non_linear_2d", position);
          REQUIRE(found);
          REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

          FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

          FaceValue<DataType> face_value{mesh_2d->connectivity()};
          parallel_for(
            face_value.numberOfItems(), PUGS_LAMBDA(const FaceId face_id) {
              const TinyVector<Dimension>& x = xl[face_id];
              face_value[face_id]            = DataType{2 * std::exp(x[0]), -3 * x[1]};
            });

          ItemValueVariantFunctionInterpoler interpoler(mesh_2d, ItemType::face, function_symbol_id);
          std::shared_ptr item_value_variant = interpoler.interpolate();

          REQUIRE(same_item_value(face_value, item_value_variant->get<FaceValue<DataType>>()));
        }

        SECTION("R3x3_non_linear_2d")
        {
          using DataType = TinyMatrix<3>;

          auto [i_symbol, found] = symbol_table->find("R3x3_non_linear_2d", position);
          REQUIRE(found);
          REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

          FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

          FaceValue<DataType> face_value{mesh_2d->connectivity()};
          parallel_for(
            face_value.numberOfItems(), PUGS_LAMBDA(const FaceId face_id) {
              const TinyVector<Dimension>& x = xl[face_id];

              face_value[face_id] = DataType{2 * std::exp(x[0]) * std::sin(x[1]) + 3,
                                             std::sin(x[1] - 2 * x[0]),
                                             3,
                                             x[1] * x[0],
                                             -4 * x[1],
                                             2 * x[0] + 1,
                                             3,
                                             -6 * x[0],
                                             std::exp(x[1])};
            });

          ItemValueVariantFunctionInterpoler interpoler(mesh_2d, ItemType::face, function_symbol_id);
          std::shared_ptr item_value_variant = interpoler.interpolate();

          REQUIRE(same_item_value(face_value, item_value_variant->get<FaceValue<DataType>>()));
        }
      }
    }
  }

  SECTION("3D")
  {
    constexpr size_t Dimension = 3;

    std::array mesh_list = MeshDataBaseForTests::get().all3DMeshes();

    for (const auto& named_mesh : mesh_list) {
      SECTION(named_mesh.name())
      {
        auto mesh_3d = named_mesh.mesh();

        auto xe = MeshDataManager::instance().getMeshData(*mesh_3d).xe();

        std::string_view data = R"(
import math;
let R_scalar_non_linear_3d: R^3 -> R, x -> 2 * exp(x[0]+x[2]) + 3 * x[1];
let R3_non_linear_3d: R^3 -> R^3, x -> [2 * exp(x[0]) + 3, x[1] - 2, 3 * x[2]];
let R2x2_non_linear_3d: R^3 -> R^2x2, x -> [[2 * exp(x[0]) * sin(x[1]) + 3, sin(x[2] - 2 * x[0])], [3, x[1] * x[0] - x[2]]];
)";
        TAO_PEGTL_NAMESPACE::string_input input{data, "test.pgs"};

        auto ast = ASTBuilder::build(input);

        ASTModulesImporter{*ast};
        ASTNodeTypeCleaner<language::import_instruction>{*ast};

        ASTSymbolTableBuilder{*ast};
        ASTNodeDataTypeBuilder{*ast};

        ASTNodeTypeCleaner<language::var_declaration>{*ast};
        ASTNodeTypeCleaner<language::fct_declaration>{*ast};
        ASTNodeExpressionBuilder{*ast};

        std::shared_ptr<SymbolTable> symbol_table = ast->m_symbol_table;

        // ensure that variables are declared at this point
        TAO_PEGTL_NAMESPACE::position position{data.size(), 1, 1, "fixture"};

        SECTION("R_scalar_non_linear_3d")
        {
          auto [i_symbol, found] = symbol_table->find("R_scalar_non_linear_3d", position);
          REQUIRE(found);
          REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

          FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

          EdgeValue<double> edge_value{mesh_3d->connectivity()};
          parallel_for(
            edge_value.numberOfItems(), PUGS_LAMBDA(const EdgeId edge_id) {
              const TinyVector<Dimension>& x = xe[edge_id];
              edge_value[edge_id]            = 2 * std::exp(x[0] + x[2]) + 3 * x[1];
            });

          ItemValueVariantFunctionInterpoler interpoler(mesh_3d, ItemType::edge, function_symbol_id);
          std::shared_ptr item_value_variant = interpoler.interpolate();

          REQUIRE(same_item_value(edge_value, item_value_variant->get<EdgeValue<double>>()));
        }

        SECTION("R3_non_linear_3d")
        {
          using DataType = TinyVector<3>;

          auto [i_symbol, found] = symbol_table->find("R3_non_linear_3d", position);
          REQUIRE(found);
          REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

          FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

          EdgeValue<DataType> edge_value{mesh_3d->connectivity()};
          parallel_for(
            edge_value.numberOfItems(), PUGS_LAMBDA(const EdgeId edge_id) {
              const TinyVector<Dimension>& x = xe[edge_id];
              edge_value[edge_id]            = DataType{2 * std::exp(x[0]) + 3, x[1] - 2, 3 * x[2]};
            });

          ItemValueVariantFunctionInterpoler interpoler(mesh_3d, ItemType::edge, function_symbol_id);
          std::shared_ptr item_value_variant = interpoler.interpolate();

          REQUIRE(same_item_value(edge_value, item_value_variant->get<EdgeValue<DataType>>()));
        }

        SECTION("R2x2_non_linear_3d")
        {
          using DataType = TinyMatrix<2>;

          auto [i_symbol, found] = symbol_table->find("R2x2_non_linear_3d", position);
          REQUIRE(found);
          REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

          FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

          EdgeValue<DataType> edge_value{mesh_3d->connectivity()};
          parallel_for(
            edge_value.numberOfItems(), PUGS_LAMBDA(const EdgeId edge_id) {
              const TinyVector<Dimension>& x = xe[edge_id];
              edge_value[edge_id] =
                DataType{2 * std::exp(x[0]) * std::sin(x[1]) + 3, std::sin(x[2] - 2 * x[0]), 3, x[1] * x[0] - x[2]};
            });

          ItemValueVariantFunctionInterpoler interpoler(mesh_3d, ItemType::edge, function_symbol_id);
          std::shared_ptr item_value_variant = interpoler.interpolate();

          REQUIRE(same_item_value(edge_value, item_value_variant->get<EdgeValue<DataType>>()));
        }
      }
    }
  }

  ASTNode::setStackDetails(stack_details);
}
